#ifndef CAMERA_H
#define CAMERA_H

#include "geometricObject.h"
#include <string>
#include <fstream>
using namespace std;

// enum XYZ_AXIS {X=0, Y=1, Z=2};	//Defined in "geometricObject.h"
// enum COLOR_AXIS {R=0, G=1, B=2};	//Defined in "geometricObject.h"



class camera
{
  public:
	  enum PROJECTION_STYLE {OTHO=0, PERSPECTIVE=1};

  private:
	PROJECTION_STYLE currentProjectionStyle;
	string cameraName;

	//*********************************************************
	//The geometric locations of the camera, its looked-at point,
	//	and the up-vector of the camera
	//*********************************************************
	float cameraCoordinate[3];
	float lookedAtPointCoordinate[3];
	float upVectorForGluLookAt[3];

	bool isCameraLookigAtFixedPoint;
	bool isCameraLookigAtFixedDirection;

	//*********************************************************
	//The following are used to set up OPENGL projection matrix
	//*********************************************************
	float left, right, bottom, top, near, far;
	float fov_degree, aspectRatio; 

	//*********************************************************
	//The following are for automatic revolving around the looked-at point
	//*********************************************************
	bool isCameraAutomaticRevolving;

	bool isPhiGoUp;
	float previousCameraCoordinate[3];
	float cameraDistanceToLookedAtPoint;
	float phi, theta;


	//*********************************************************
	//The following are for flight simualtion using this camera
	//*********************************************************
	bool isCameraInFlightSimulation;

	//Set the boundaries of the flight Zone 
	float flightMaxX;
	float flightMinX;
	float flightMaxY;
	float flightMinY;
	float flightMaxZ;
	float flightMinZ;

	float flightSpeed;
	float unitSpeedVector[3];

	float headPullUpAngle;  //in radius
	float bodyRotateAngle;  //in radius

	float previousLookedAtPointCoordinate[3];


public:

	camera();

	void setCameraNameAndProjectionStyle( string name, PROJECTION_STYLE style);

	void resetCameraPosition();

	void useDefaultOthoProjection();

	void useDefaultPerspectiveProjection();

	void do_OpenGL_LookAt_BasedOnThisCamera();

	void saveToFile_OpenGL_LookAt_BasedOnThisCamera(ofstream & outfile);

	void do_OpenGL_Projection_BasedOnThisCamera(int w, int h);

	void saveToFile_OpenGL_Projection_BasedOnThisCamera(ofstream  & outfile);

	//Get the raster position for showing text message on the top of the view
	void getRasterPositionLeftTopCorner(float& xPos, float& yPos, float& zPos);

	//Get the raster position for showing text message on the bottom of the view
	void getRasterPositionLeftBottomCorner(float& xPos, float& yPos, float& zPos);

	void setCameraCoordinates(float x, float y, float z)
		{	cameraCoordinate[X] = x; 
			cameraCoordinate[Y] = y; 
			cameraCoordinate[Z] = z;	
		};

	void getCameraCoordinates(float &x, float &y, float &z)
		{	x=cameraCoordinate[X]; 
			y=cameraCoordinate[Y]; 
			z=cameraCoordinate[Z];
		};

	void setLookedAtPointCoordinates(float x, float y, float z)
		{	lookedAtPointCoordinate[X] = x; 
			lookedAtPointCoordinate[Y] = y; 
			lookedAtPointCoordinate[Z] = z;
		};

	void getLookedAtPointCoordinates(float &x, float &y, float &z)
		{	x=lookedAtPointCoordinate[X]; 
			y=lookedAtPointCoordinate[Y]; 
			z=lookedAtPointCoordinate[Z];
		};

	void setUpVectorForGluLookAt(float x, float y, float z)
		{	upVectorForGluLookAt[X]= x; 
			upVectorForGluLookAt[Y]= y; 
			upVectorForGluLookAt[Z]= z;
		};

	void getUpVectorForGluLookAt(float &x, float &y, float &z)
		{	x=upVectorForGluLookAt[X]; 
			y=upVectorForGluLookAt[Y]; 
			z=upVectorForGluLookAt[Z];
		};

	void flipCameraToOtherSideOfAxis(XYZ_AXIS axis)
	{
		cameraCoordinate[axis] *= -1.0;
		isCameraLookigAtFixedPoint = true;
		isCameraLookigAtFixedDirection = false;

	}

	void moveCameraAlongAxis(XYZ_AXIS axis, float units)
	{	cameraCoordinate[axis] += units;
		if (!isCameraLookigAtFixedPoint)
			lookedAtPointCoordinate[axis] += units;
	};

	void zoomInCameraView();

	void zoomOutCameraView();

	void turnCameraToLookAtThisPoint(float x, float y, float z)
	{	lookedAtPointCoordinate[X] = x;
		lookedAtPointCoordinate[Y] = y;
		lookedAtPointCoordinate[Z] = z;
	};

	void keepCameraLookingAtTheSamePoint()
		{	isCameraLookigAtFixedPoint = true;
			isCameraLookigAtFixedDirection = false;
		};

	void keepCameraLookingAtTheSameDirection()
		{	isCameraLookigAtFixedDirection = true;
			isCameraLookigAtFixedPoint = false;
		};

	void moveLookedAtPointAlongAxis(XYZ_AXIS axis, float units)
		{lookedAtPointCoordinate[axis] += units;} ;



	void startCameraAutomaticRevolvingAroundTheLookedAtPoint();

	void setupAutomaticRevolving();

	void stopCameraAutomaticRevolvingAroundTheLookedAtPoint_ReturnToPreviousLocation();

	void stopCameraAutomaticRevolvingAroundTheLookedAtPoint_StayHere();

	void revolveOneStepAroundTheLookedAtPoint();

	bool isCameraInAutomaticRevolvingMode()
	{return isCameraAutomaticRevolving;};



	void startFlightSimulation
	(float maxX, float minX, float maxY, float minY, float maxZ, float minZ);

	void stopFlightSimulation_ReturnToPreviousLocation();

	void stopFlightSimulation_StayHere();

	void flyOneStepInFlightSimulation();

	void updateFlightSimulationSpeedVector();

	void updateFlightSimulationUpVectorForGluLookAt();

	void setFlySimulationParametersAccordingToMouseMotion
		(double mouseMotionX, double  mouseMotionY);

	bool isCameraInFlightSimulationMode()
	{return isCameraInFlightSimulation;};


	static void normalizeIntoUnitVector(float * aVector, int size);


	void displayInfo();

	void saveToFile_CameraInfo(ofstream & outfile);

	void loadFromFile_CameraInfo(ifstream & infile);

	bool projectToScreenCoordinates
	(	int windowWidth,	int windowHeight,
		double x,		double y,		double z,
		double& winX,	double& winY,	double& winZ
	);

	//Map the mouse motion into movement of camera location
	void moveCameraAccordingToMouseMotion(double mouseMotionX, double  mouseMotionY);
		
	//The following are default constants used to set up OPENGL projection matrix
	static const float MyLEFT;		
	static const float MyRIGHT;
	static const float MyBOTTOM;	
	static const float MyTOP;
	static const float MyNEAR;	
	static const float MyFAR;
	static const float My_FOV_DEGREE;
	static const float My_ASPECT_RATIO; 

	//The following are constants measuring a basic unit of change of view
	static const float UNIT_PERSPECTIVE_FOV_CHANGE;	//1 degree
	static const float UNIT_OTHO_CHANGE	;		//0.1 unit length

	//THe default distance of the camera from the origin
	static const float DEFAULT_DISTANCE_FROM_ORIGIN;

	//A speed factor to adjust how fast the camera can revolve around 
	static const int SPEED_FACTOR;
};

#endif
